/* TODO:
 *   - generalize cosmology file parsing to a general-purpose configuration file
 *     parser; lines of the form "variable = value" fill a dictionary; more general
 *     behavior is achieved by subclassing the ConfigurationFile class
 *   - change file extension for cosmology files? */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <cstdio>
#include <sys/stat.h>

#include "Cosmology.h"
#include "Datafile.h"
#include "Math.h"
#include "pstring.h"


Cosmology::Cosmology() {
    /* No initialization */
}

Cosmology::Cosmology(real h, real Tcmb, real n, real Omega_m, real Omega_b) {
    Initialize(h, Tcmb, n, Omega_m, Omega_b);
}

Cosmology::Cosmology(real h, real Tcmb, real n, real Omega_m, real Omega_b, real sigma8, const array& ki, const array& Ti) {
    Initialize(h, Tcmb, n, Omega_m, Omega_b);
    SetTransferFunction(ki, Ti);
    NormalizeTransferFunction(sigma8);
}

Cosmology::Cosmology(const char* name) {
    sigma8 = 0;
    delta_H = 1;
    Initialize(name);
}

void Cosmology::Initialize(real h_, real Tcmb_, real n_, real Omega_m_, real Omega_b_) {
    h = h_;
    Tcmb = Tcmb_;
    n = n_;
    Omega_m = Omega_m_;
    Omega_b = Omega_b_;

    sigma8 = 0;
    delta_H = 1;

    CalculateAdditionalParameters();
}

bool is_regular_file(const char* path) {
    struct stat buf;
    int error = stat(path, &buf);
    return S_ISREG(buf.st_mode);
}

void Cosmology::Initialize(const char* tag) {
    /* First check the current directory for a file named 'tag' */
    pstring path(tag);
    FILE* fp = fopen(path.c_str(), "r");

    if(!fp || !is_regular_file(path.c_str())) {
        /* Next try adding a file extension to 'tag' */
        path += ".ini";
        fp = fopen(path.c_str(), "r");
    }

    if(!fp || !is_regular_file(path.c_str())) {
        /* Finally search in the default data directory */
        path = DATADIR "/" + path;
        fp = fopen(path.c_str(), "r");
    }

    if(!fp || !is_regular_file(path.c_str())) {
        fprintf(stderr, "Cosmology: could not find cosmology '%s'\n", tag);
        fprintf(stderr, "  tried ./%s, ./%s.ini, and %s/%s.ini\n", tag, tag, DATADIR, tag);
        return;
    }

    verbose("Reading cosmology from %s\n", path.c_str());

    char buf[1024];
    while(fgets(buf, sizeof(buf), fp) != NULL) {
        pstring line(buf);
        line.strip(true);
        if(line.empty() || line.startswith('#'))
            continue;
        pstring::size_type j = line.find('=');
        if(j == pstring::npos) {
            fprintf(stderr, "invalid entry: %s\n", line.c_str());
            continue;
        }

        pstring var = pstring(line, 0, j).strip();
        pstring value = pstring(line, j+1).strip();
        if(var == "h")
            h = (double)value;
        else if(var == "Tcmb")
            Tcmb = (double)value;
        else if(var == "n")
            n = (double)value;
        else if(var == "Omega_m")
            Omega_m = (double)value;
        else if(var == "Omega_b")
            Omega_b = (double)value;
        else if(var == "sigma8")
            sigma8 = (double)value;
        else if(var == "tkfile") {
            vector<pstring> fields = value.split(':');
            pstring tkfile = fields[0];
            int kcol = (fields.size() >= 2) ? int(fields[1]) : 1;
            int tcol = (fields.size() >= 3) ? int(fields[2]) : 2;
            LoadTransferFunction(tkfile.c_str(), kcol, tcol);
        }
    }
    fclose(fp);

    CalculateAdditionalParameters();
    if(!Ti.empty() && sigma8 != 0)
        NormalizeTransferFunction(sigma8);
}

void Cosmology::SetTransferFunction(const array& ki_, const array& Ti_) {
    ki = ki_;
    Ti = Ti_;
}

void Cosmology::LoadTransferFunction(const char* tkfile, int kcol, int tcol) {
    /* First check the current directory for a file named 'tkfile' */
    pstring tkpath(tkfile);
    FILE* fp = fopen(tkpath.c_str(), "r");

    if(!fp) {
        /* Next search in the default data directory */
        tkpath = DATADIR "/" + tkpath;
        fp = fopen(tkpath.c_str(), "r");
    }

    if(!fp) {
        fprintf(stderr, "Cosmology: could not find transfer function '%s'\n", tkfile);
        fprintf(stderr, "  tried ./%s and %s/%s\n", tkfile, DATADIR, tkfile);
        return;
    }

    verbose("Reading transfer function from %s\n", tkpath.c_str());

    Datafile data(fp);
    ki = data.GetColumn(kcol);
    Ti = data.GetColumn(tcol);
}

void Cosmology::NormalizeTransferFunction(real sigma8) {
    /* Calculate sigma8 directly from the transfer function (with delta_H = 1) */
    real R = 8;
    int N = ki.size() - 1;
    array fi(N+1);
    for(int i = 0; i <= N; i++) {
        real k = ki[i];
        fi[i] = pow(2997.925*k, 3+n) * pow2(Ti[i]) * pow2(3*SpecialFunctions::j_1(k*R)/(k*R));
    }
    real S = 0;
    for(int i = 0; i <= N-1; i++)
        S += log(ki[i+1]/ki[i]) * 0.5*(fi[i] + fi[i+1]);        // trapezoid rule in log(k)

    /* Set delta_H to get the desired sigma8 */
    delta_H = sigma8/sqrt(S);
}

void Cosmology::CalculateAdditionalParameters() {
    using namespace SIUnits;
    H0 = 100*h * km/second/Mpc;
    rho_crit = (3*H0*H0)/(8*M_PI*G);
    Omega_gamma = (M_PI*M_PI/15) * pow4(k*Tcmb)/(pow3(hbar)*pow5(c)) / rho_crit;
    Omega_nu = 3.*(7./8.) * pow(4./11., 4./3.) * Omega_gamma;
    Omega_r = Omega_gamma + Omega_nu;
    Omega_Lambda = 1 - Omega_m;
    a_eq = Omega_r/Omega_m;
    z_eq = 1/a_eq - 1;
    k_eq = (H0/c) * sqrt(2*Omega_m/a_eq) / (h/Mpc);
}

void Cosmology::Print(const char* prefix) const {
    printf("%sh = %g\n", prefix, h);
    printf("%sTcmb = %g\n", prefix, Tcmb);
    printf("%sn = %g\n", prefix, n);
    printf("%sOmega_m = %g\n", prefix, Omega_m);
    printf("%sOmega_b = %g\n", prefix, Omega_b);
    printf("%ssigma8 = %g\n", prefix, sigma8);
    printf("%sH0 = %g\n", prefix, H0);
    printf("%srho_crit = %g\n", prefix, rho_crit);
    printf("%sOmega_gamma = %g\n", prefix, Omega_gamma);
    printf("%sOmega_nu = %g\n", prefix, Omega_nu);
    printf("%sOmega_r = %g\n", prefix, Omega_r);
    printf("%sOmega_Lambda = %g\n", prefix, Omega_Lambda);
    printf("%sa_eq = %g\n", prefix, a_eq);
    printf("%sz_eq = %g\n", prefix, z_eq);
    printf("%sk_eq = %g\n", prefix, k_eq);
    printf("%sdelta_H = %g\n", prefix, delta_H);
}
