#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <cassert>
#include <cstdio>

#include "Math.h"
#include "Spline.h"


struct SplineImpl {
    int N;
    const array& X;
    const array& Y;

    SplineImpl(int nn, const array& xx, const array& yy)
        : N(nn), X(xx), Y(yy) {}
    virtual ~SplineImpl() {}

    virtual real y(int i, real x) const = 0;
    virtual real dydx(int i, real x) const = 0;
//    virtual real max(int i, real& xmax) const = 0;
    virtual SplineImpl* clone(const array& newX, const array& newY) const = 0;
};


/*****************
 * Linear spline *
 *****************/

struct LinearSplineImpl : public SplineImpl {
    LinearSplineImpl(int N, const array& X, const array& Y) : SplineImpl(N, X, Y) {
    }

    real y(int i, real x) const {
        assert(0 <= i && i <= N-2);
        real x1 = X[i], x2 = X[i+1], y1 = Y[i], y2 = Y[i+1];
        return y1 + (y2 - y1)*(x - x1)/(x2 - x1);
    }

    real dydx(int i, real x) const {
        real x1 = X[i], x2 = X[i+1], y1 = Y[i], y2 = Y[i+1];
        return (y2 - y1)/(x2 - x1);
    }

    real max(int i, real& xmax) const {
        real x1 = X[i], x2 = X[i+1], y1 = Y[i], y2 = Y[i+1];
        if(y1 > y2) {
            xmax = x1;
            return y1;
        }
        else {
            xmax = x2;
            return y2;
        }
    }

    LinearSplineImpl* clone(const array& newX, const array& newY) const {
        return new LinearSplineImpl(N, newX, newY);
    }
};


/************************************************************
 * Cubic spline
 * 
 * Implementation copied from netlib's fmm/spline.f program.
 ************************************************************/

struct CubicSplineImpl : public SplineImpl {
    /* y(i,x) = a + b (x-xi) + c (x-xi)^2 + d (x-xi)^3 */
    array b, c, d;

    CubicSplineImpl(int N, const array& X, const array& Y) : SplineImpl(N, X, Y) {
        assert(N >= 2);
        b.resize(N);
        c.resize(N);
        d.resize(N);

        if(N == 2) {
            b[0] = b[1] = (Y[1] - Y[0])/(X[1] - X[0]);
            c[0] = c[1] = 0;
            d[0] = d[1] = 0;
        }
        else {
            /* Set up tridiagonal system: b = diagonal, d = offdiagonal, c = right hand side */
            d[0] = X[1] - X[0];
            c[1] = (Y[1] - Y[0])/d[0];
            for(int i = 1; i < N-1; i++) {
                d[i] = X[i+1] - X[i];
                b[i] = 2*(d[i-1] + d[i]);
                c[i+1] = (Y[i+1] - Y[i])/d[i];
                c[i] = c[i+1] - c[i];
            }

            /* End conditions: third derivatives at X[0] and X[N-1] obtained from divide differences */
            b[0] = -d[0];
            b[N-1] = -d[N-2];
            c[0] = 0;
            c[N-1] = 0;
            if(N > 3) {
                c[0] = c[2]/(X[3] - X[1]) - c[1]/(X[2] - X[0]);
                c[N-1] = c[N-2]/(X[N-1] - X[N-3]) - c[N-3]/(X[N-2] - X[N-4]);
                c[0] = c[0]*d[0]*d[0]/(X[3] - X[0]);
                c[N-1] = -c[N-1]*d[N-1]*d[N-1]/(X[N-1] - X[N-4]);
            }

            /* Forward elimination */
            for(int i = 1; i < N; i++) {
                real t = d[i-1]/b[i-1];
                b[i] -= t*d[i-1];
                c[i] -= t*c[i-1];
            }

            /* Back substitution */
            c[N-1] = c[N-1]/b[N-1];
            for(int j = 0; j < N-1; j++) {
                int i = N - j - 2;
                c[i] = (c[i] - d[i]*c[i+1])/b[i];
            }

            /* Compute polynomial coefficients */
            b[N-1] = (Y[N-1] - Y[N-2])/d[N-2] + d[N-2]*(c[N-2] + 2*c[N-1]);
            for(int i = 0; i < N-1; i++) {
                b[i] = (Y[i+1] - Y[i])/d[i] - d[i]*(c[i+1] + 2*c[i]);
                d[i] = (c[i+1] - c[i])/d[i];
                c[i] = 3*c[i];
            }
            c[N-1] = 3*c[N-1];
            d[N-1] = d[N-2];
        }
    }

    CubicSplineImpl(int N, const array& X, const array& Y,
                    const array& bb, const array& cc, const array& dd)
        : SplineImpl(N, X, Y)
    {
        b = bb;
        c = cc;
        d = dd;
    }

    real y(int i, real x) const {
        real u = x - X[i];
        return Y[i] + b[i]*u + c[i]*u*u + d[i]*u*u*u;
    }

    real dydx(int i, real x) const {
        real u = x - X[i];
        return b[i] + 2*c[i]*u + 3*d[i]*u*u;
    }

//    real max(int i, real& xmax) const {
//        return 0;
//    }

    CubicSplineImpl* clone(const array& newX, const array& newY) const {
        return new CubicSplineImpl(N, newX, newY, b, c, d);
    }
};


/**********
 * Spline *
 **********/

Spline::Spline() {
    method = -1;
    N = 0;
    impl = 0;
}

Spline::Spline(const array& xx, const array& yy, int m) {
    N = (int)xx.size();
    assert(N >= 2);
    X = xx;
    Y = yy;
    method = m;
    if(method == Linear)
        impl = new LinearSplineImpl(N, X, Y);
    else if(method == Cubic)
        impl = new CubicSplineImpl(N, X, Y);
    else
        impl = 0;
}

Spline::Spline(int nn, const real* xx, const real* yy, int m) {
    N = nn;
    assert(N >= 2);
    X.setdata(N, xx);
    Y.setdata(N, yy);
    method = m;
    if(method == Linear)
        impl = new LinearSplineImpl(N, X, Y);
    else if(method == Cubic)
        impl = new CubicSplineImpl(N, X, Y);
    else
        impl = 0;
}

Spline::Spline(const Spline& S) {
    method = S.method;
    X = S.X;
    Y = S.Y;
    N = S.N;
    if(S.impl)
        impl = S.impl->clone(X, Y);
    else
        impl = 0;
}

Spline::~Spline() {
    delete impl;
}

Spline& Spline::operator=(const Spline& S) {
    delete impl;
    method = S.method;
    X = S.X;
    Y = S.Y;
    N = S.N;
    impl = S.impl->clone(X, Y);

    return *this;
}

real Spline::Evaluate(real x) const {
    if(N < 2 || x < X[0] || x > X[N-1])
        return 0;
    else {
        int i = LookupIndex(x);
        return impl->y(i, x);
    }
}

real Spline::EvaluateDerivative(real x) const {
    if(N < 2 || x < X[0] || x > X[N-1])
        return 0;
    else {
        int i = LookupIndex(x);
        return impl->dydx(i, x);
    }
}

#if 0
real InterpolatingFunction::FindMaximum(real xguess, real* yret) {
    assert(N >= 0);

    real xmax, ymax;
    if(xguess < X[0]) {
        ymax = yleft;
        xmax = xguess;
    }
    else if(xguess > X[N]) {
        ymax = yright;
        xmax = xguess;
    }
    else {
        /* Now we walk from one segment to the next until we find a maximum */

        int i = LookupIndex(xguess);
        /* If we're right on a boundary the algorithm breaks */
        if(xguess == X[i])
            xguess += (X[i+1]-X[i])/1000;
        else if(xguess == X[i+1])
            xguess -= (X[i+1]-X[i])/1000;

        xmax = GetSegment(i).max(ymax);
        if(xmax == xguess)
            xmax += (X[i+1]-X[i])/100;
        int dir = (xmax > xguess) ? +1 : -1;

        while(true) {
            if(X[i] < xmax && xmax < X[i+1])    // local maximum
                break;
            else if(xmax == X[i]) {     // maximum is the left-most point of the interval
                if(dir == +1)           // if we started off walking right, we're done
                    break;
                if(i == 0)              // if we can't go any further left, we're done
                    break;
                i--;
            }
            else if(xmax == X[i+1]) {   // maximum is the right-most point of the interval
                if(dir == -1)           // if we started off walking left, we're done
                    break;
                if(i == N-1)            // if we can't go any further right, we're done
                    break;
                i++;
            }
            else {
                fprintf(stderr, "InterpolatingFunction::FindMaximum: broken\n");
            }
            xmax = GetSegment(i).max(ymax);
        }
    }

    if(yret != NULL)
        *yret = ymax;
    return xmax;
}
#endif

int Spline::LookupIndex(real x) const {
    /* We should never get here if we're out of range */
    assert(!isinf(x) && !isnan(x) && X[0] <= x && x <= X[N-1]);

    /* Potentially pathological case */
    if(x == X[N-1])
        return N-2;

    /* First calculate what the index would be for uniformly spaced points */
    real I = (N-1) * (x - X[0])/(X[N-1] - X[0]);
    int i = (int)I;

    /* Check this index and its immediate neighbors, in case the points are
     * actually uniform (or close to it) */
    if(X[i] <= x && x <= X[i+1])
        return i;
    else if(X[i-1] <= x && x < X[i])
        return i-1;
    else if(X[i+1] < x && x <= X[i+2])
        return i+1;
    else {
        /* Finally resort to a binary search starting from the initial guess */
        int ilo = 0;
        int ihi = N-2;
        while(true) {
            if(x < X[i]) {
                ihi = i-1;
                i = (ilo+ihi)/2;
            }
            else if(x > X[i+1]) {
                ilo = i+1;
                i = (ilo+ihi)/2;
            }
            else
                break;
        }
        return i;
    }
}
